# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
import os
import sys
import importlib
import importlib.util
import tempfile
import logging
import socket
from multiprocessing import Queue, Process, set_start_method
from .configobj import ConfigObj
from . import ipsLogging
from .services import ServicesProxy
from .componentRegistry import ComponentID, ComponentRegistry

# Try using fork for starting subprocesses, this is the default on
# Linux but not macOS with python >= 3.8
if sys.platform == 'darwin':
    try:
        set_start_method('fork')
    except RuntimeError:
        # context can only be set once
        pass


class ConfigurationManager:
    """
    The configuration manager is responsible for paring the simulation and
    platform configuration files, creating the framework and simulation
    components, as well as providing an interface to accessing items from the
    configuration files (e.g., the time loop).
    """

    # CM init
    class SimulationData:
        """
        Structure to hold simulation data stored into the sim_map
        entry in the configurationManager class
        """

        def __init__(self, sim_name):
            self.sim_name = sim_name
            self.portal_sim_name = None
            self.sim_root = None
            self.sim_conf = None
            self.config_file = None
            self.conf_file_dir = None
            self.driver_comp = None
            self.init_comp = None
            self.all_comps = []
            self.port_map = {}
            self.component_process = None
            self.process_list = []

    def __init__(self, fwk, config_file_list, platform_file_name):
        """
        Initialize the values to be used by the configuration manager.  Also
        specified are the required fields of the simulation configuration
        file, and the configuration files are read in.
        """
        # ref to framework
        self.fwk = fwk
        self.event_mgr = None
        self.data_mgr = None
        self.resource_mgr = None
        self.task_mgr = None
        self.comp_registry = ComponentRegistry()
        # SIMYAN: here is where we removed the requirement for BIN_PATH, etc.
        # from the required fields. This was done so that we could specify it
        # in the component-generic.conf file, which allows you to point to a
        # directory that contains physics and other binaries on a global level
        # i.e. removing the requirement that it be specified for each component
        self.required_fields = set(['CLASS', 'SUB_CLASS', 'NAME', 'SCRIPT',
                                    'INPUT_FILES', 'OUTPUT_FILES', 'NPROC'])
        self.config_file_list = []
        self.sim_name_list = None
        self.sim_root_list = None
        self.log_file_list = None
        self.log_dynamic_sim_queue = Queue(0)

        class Unbuffered:
            def __init__(self, stream):
                self.stream = stream

            def write(self, data):
                self.stream.write(data)
                self.stream.flush()

            def writelines(self, datas):
                self.stream.writelines(datas)
                self.stream.flush()

            def __getattr__(self, attr):
                return getattr(self.stream, attr)

        for conf_file in config_file_list:
            abs_path = os.path.abspath(conf_file)
            if abs_path not in self.config_file_list:
                self.config_file_list.append(abs_path)
            else:
                print('Ignoring duplicate configuration file ', abs_path)

        # sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
        sys.stdout = Unbuffered(sys.stdout)
        self.platform_file = os.path.abspath(platform_file_name)
        self.platform_conf = {}
        loc_keys = []
        mach_keys = ['MPIRUN', 'NODE_DETECTION', 'CORES_PER_NODE', 'SOCKETS_PER_NODE', 'NODE_ALLOCATION_MODE']
        prov_keys = ['HOST']
        self.platform_keywords = loc_keys + mach_keys + prov_keys

        self.service_methods = ['get_port',
                                'get_config_parameter',
                                'set_config_parameter',
                                'get_time_loop',
                                'create_simulation']
        self.fwk.register_service_handler(self.service_methods,
                                          getattr(self, 'process_service_request'))
        self.sim_map = {}
        self.finished_sim_map = {}
        self.fwk_sim_name = None  # "Fake" simconf for framework components
        self.fwk_components = []  # List of framework specific components
        self.myTopic = None
        self.log_daemon = ipsLogging.ipsLogger(self.log_dynamic_sim_queue)
        self.log_process = None

    # CM initialize
    def initialize(self, data_mgr, resource_mgr, task_mgr):
        """
        Parse the platform and simulation configuration files using the
        :py:obj:`ConfigObj` module.  Create and initialize simulation(s) and
        their components, framework components and loggers.
        """
        self.event_mgr = None  # eventManager(self)
        self.data_mgr = data_mgr
        self.resource_mgr = resource_mgr
        self.task_mgr = task_mgr
        # Parse configuration files into configuration map
        sim_root_list = self.sim_root_list = []
        sim_name_list = self.sim_name_list = []
        log_file_list = self.log_file_list = []

        # Idiot checks
        if len(self.config_file_list) == 0:
            self.fwk.exception('Missing config file? Something is very wrong')
            raise ValueError('Missing config file? Something is very wrong')

        """
        Platform Configuration
        """
        # parse file
        try:
            self.platform_conf = ConfigObj(self.platform_file,
                                           interpolation='template',
                                           file_error=True)
        except (IOError, SyntaxError):
            self.fwk.exception('Error opening config file: %s',
                               self.platform_file)
            raise
        # get mandatory values
        for kw in self.platform_keywords:
            try:
                self.platform_conf[kw]
            except KeyError:
                self.fwk.exception('Missing required parameter %s in platform config file',
                                   kw)
                raise
        # Make sure the HOST variable is defined
        try:
            host = self.platform_conf['HOST']
        except KeyError:
            self.platform_conf['HOST'] = socket.gethostname()
        else:
            if not host:
                self.platform_conf['HOST'] = socket.gethostname()

        """
        optional platform values are obtained and read here
        """
        user = ''
        try:
            user = self.platform_conf['USER']
        except KeyError:
            try:
                user = os.environ['USER']
            except Exception:
                pass
        self.platform_conf['USER'] = user

        # Grab environment variables
        plat_keys = list(self.platform_conf.keys())
        for (k, v) in os.environ.items():
            if k not in plat_keys \
                    and not any(x in v for x in '{}()$'):
                self.platform_conf[k] = v

        mpirun_version = self.platform_conf.get('MPIRUN_VERSION', 'OpenMPI-generic')

        # node allocation mode describes how node allocation should be handled
        # in the IPS.
        #  EXCLUSIVE - only one application can run on a single node.
        #  SHARE - applications may share nodes.

        try:
            node_alloc_mode = self.platform_conf['NODE_ALLOCATION_MODE'].upper()
            if node_alloc_mode not in ['EXCLUSIVE', 'SHARED']:
                self.fwk.exception("bad value for NODE_ALLOCATION_MODE. expected 'EXCLUSIVE' or 'SHARED'.")
                raise ValueError("bad value for NODE_ALLOCATION_MODE. expected 'EXCLUSIVE' or 'SHARED'.")
        except Exception:
            self.fwk.exception("missing value or bad type for NODE_ALLOCATION_MODE.  expected 'EXCLUSIVE' or 'SHARED'.")
            raise

        uan_val = self.platform_conf.get('USE_ACCURATE_NODES', 'ON').upper()
        if uan_val in ['OFF', 'FALSE']:
            use_accurate_nodes = False
        else:
            use_accurate_nodes = True

        self.platform_conf['TOTAL_PROCS'] = int(self.platform_conf.get('TOTAL_PROCS', 0))
        self.platform_conf['NODES'] = int(self.platform_conf.get('NODES', 0))
        self.platform_conf['PROCS_PER_NODE'] = int(self.platform_conf.get('PROCS_PER_NODE', 0))
        self.platform_conf['CORES_PER_NODE'] = int(self.platform_conf.get('CORES_PER_NODE', 0))
        self.platform_conf['SOCKETS_PER_NODE'] = int(self.platform_conf.get('SOCKETS_PER_NODE', 0))
        self.platform_conf['USE_ACCURATE_NODES'] = use_accurate_nodes
        self.platform_conf['MPIRUN_VERSION'] = mpirun_version

        """
        Simulation Configuration
        """
        for conf_file in self.config_file_list:
            try:
                conf = ConfigObj(conf_file, interpolation='template', file_error=True)

                # Import environment variables into config file
                # giving precedence to config file definitions in case of duplicates
                conf_keys = list(conf.keys())
                for (k, v) in os.environ.items():
                    if k not in conf_keys and not any(x in v for x in '{}()$'):
                        conf[k] = v

                # Allow simulation file to override platform values
                # and then put all platform values into simulation map
                for key in self.platform_conf:
                    if key in conf_keys and key not in os.environ.keys():
                        self.platform_conf[key] = conf[key]
                    if key not in conf_keys:
                        conf[key] = self.platform_conf[key]

                # Override platform value for PORTAL_URL if in simulation
                if 'PORTAL_URL' in conf_keys:
                    self.platform_conf['PORTAL_URL'] = conf['PORTAL_URL']

            except (IOError, SyntaxError):
                self.fwk.exception('Error opening config file %s: ', conf_file)
                raise
            except Exception:
                self.fwk.exception('Error(s) during parsing of supplied config file %s: ', conf_file)
                raise

            try:
                sim_name = conf['SIM_NAME']
                sim_root = conf['SIM_ROOT']
                log_file = os.path.abspath(conf['LOG_FILE'])
            except KeyError:
                self.fwk.exception('Missing required parameters SIM_NAME, SIM_ROOT or LOG_FILE\
 in configuration file %s', conf_file)
                raise

            if sim_name in sim_name_list:
                self.fwk.exception('Error: Duplicate SIM_NAME in configuration files')
                sys.exit(1)
            if sim_root in sim_root_list:
                self.fwk.exception('Error: Duplicate SIM_ROOT in configuration files')
                sys.exit(1)
            if log_file in log_file_list:
                self.fwk.exception('Error: Duplicate LOG_FILE in configuration files')
                sys.exit(1)
            if 'SIMULATION_CONFIG_FILE' not in conf:
                conf['SIMULATION_CONFIG_FILE'] = conf_file
            sim_name_list.append(sim_name)
            sim_root_list.append(sim_root)
            log_file_list.append(log_file)
            new_sim = self.SimulationData(sim_name)
            conf['__PORTAL_SIM_NAME'] = sim_name
            new_sim.sim_conf = conf
            new_sim.config_file = conf_file
            new_sim.portal_sim_name = sim_name

            # SIMYAN: store the directory of the configuration file
            new_sim.conf_file_dir = os.path.dirname(os.path.abspath(conf_file))
            new_sim.sim_root = sim_root
            new_sim.log_file = log_file
            new_sim.log_pipe_name = tempfile.mktemp('.logpipe', 'ips_')

            self.log_daemon.add_sim_log(new_sim.log_pipe_name,
                                        new_sim.log_file)
            self.sim_map[sim_name] = new_sim

            # Use first simulation for framework components
            if not self.fwk_sim_name:
                fwk_sim_conf = conf.dict()
                fwk_sim_conf['SIM_NAME'] = '_'.join([conf['SIM_NAME'], 'FWK'])
                fwk_sim = self.SimulationData(fwk_sim_conf['SIM_NAME'])
                fwk_sim.sim_conf = fwk_sim_conf
                fwk_sim.sim_root = new_sim.sim_root
                fwk_sim.log_file = self.fwk.log_file  # sys.stdout
                fwk_sim.log_pipe_name = tempfile.mktemp('.logpipe', 'ips_')
                fwk_sim_conf['LOG_LEVEL'] = 'DEBUG'
                self.log_daemon.add_sim_log(fwk_sim.log_pipe_name, fwk_sim.log_file)
                self.fwk_sim_name = fwk_sim_conf['SIM_NAME']
                self.sim_map[fwk_sim.sim_name] = fwk_sim

        self.log_process = Process(target=self.log_daemon.__run__)
        self.log_process.start()

        for sim_name, sim_data in self.sim_map.items():
            if sim_name != self.fwk_sim_name:
                self._initialize_sim(sim_data)

        # ***** commenting out portal stuff for now
        self._initialize_fwk_components()

        # do later - subscribe to events, set up event publishing structure
        # publish "CM initialized" event

    def _initialize_fwk_components(self):
        """
        Initialize 'components' that are part of the framework infrastructure.
        Those components (for now) communicate using the event bus and are not
        part of the normal framework-mediated RPC inter-compponent interactions
        """

        # SIMYAN: set up the runspaceInit component
        runspace_conf = {}
        runspace_conf['CLASS'] = 'FWK'
        runspace_conf['SUB_CLASS'] = 'COMP'
        runspace_conf['NAME'] = 'runspaceInitComponent'
        runspace_conf['SCRIPT'] = ''
        runspace_conf['MODULE'] = 'ipsframework.runspaceInitComponent'
        runspace_conf['INPUT_DIR'] = '/dev/null'
        runspace_conf['INPUT_FILES'] = ''
        runspace_conf['IPS_CONFFILE_DIR'] = ''
        runspace_conf['DATA_FILES'] = ''
        runspace_conf['OUTPUT_FILES'] = ''
        runspace_conf['NPROC'] = 1
        runspace_conf['LOG_LEVEL'] = 'WARNING'
        runspace_conf['OS_CWD'] = os.getcwd()
        if self.fwk.log_level == logging.DEBUG:
            runspace_conf['LOG_LEVEL'] = 'DEBUG'

        runspace_component_id = self._create_component(runspace_conf,
                                                       self.sim_map[self.fwk_sim_name])
        self.fwk_components.append(runspace_component_id)

        # SIMYAN: set up The Portal bridge, allowing for an absense of a portal
        use_portal = True
        if 'USE_PORTAL' in self.sim_map[self.fwk_sim_name].sim_conf:
            use_portal = self.sim_map[self.fwk_sim_name].sim_conf['USE_PORTAL']
            if use_portal.lower() == "false":
                use_portal = False
        if use_portal:
            portal_conf = {}
            portal_conf['CLASS'] = 'FWK'
            portal_conf['SUB_CLASS'] = 'COMP'
            portal_conf['NAME'] = 'PortalBridge'
            if 'FWK_COMPS_PATH' in self.sim_map[self.fwk_sim_name].sim_conf:
                portal_conf['BIN_PATH'] = self.sim_map[self.fwk_sim_name].sim_conf['FWK_COMPS_PATH']
                portal_conf['SCRIPT'] = os.path.join(portal_conf['BIN_PATH'], 'portalBridge.py')
            else:
                portal_conf['SCRIPT'] = ''
                portal_conf['MODULE'] = 'ipsframework.portalBridge'
            portal_conf['INPUT_DIR'] = '/dev/null'
            portal_conf['INPUT_FILES'] = ''
            portal_conf['DATA_FILES'] = ''
            portal_conf['OUTPUT_FILES'] = ''
            portal_conf['NPROC'] = 1
            portal_conf['LOG_LEVEL'] = 'WARNING'
            try:
                portal_conf['USER'] = self.sim_map[self.fwk_sim_name].sim_conf['USER']
            except KeyError:
                portal_conf['USER'] = self.platform_conf['USER']
            if self.fwk.log_level == logging.DEBUG:
                portal_conf['LOG_LEVEL'] = 'DEBUG'

            portal_conf['PORTAL_URL'] = self.get_platform_parameter('PORTAL_URL', silent=True)

            component_id = self._create_component(portal_conf,
                                                  self.sim_map[self.fwk_sim_name])
            self.fwk_components.append(component_id)

    def _initialize_sim(self, sim_data):
        """
        Parses the configuration data (*sim_conf*) associated with a simulation
        (*sim_name*). Instantiate the components associated with each simulation.
        Populate the *component_registry* with appropriate component and port
        mapping info.
        """
        sim_conf = sim_data.sim_conf
        sim_name = sim_data.sim_name
        ports_config = sim_conf['PORTS']
        ports_list = ports_config['NAMES'].split()

        # simRootDir = self.get_sim_parameter(sim_name, 'SIM_ROOT')
        # SIMYAN: removed code that would make the simrootDir from here and
        # moved it to the runspaceInit component
        # set simulation level partial_nodes
        try:
            pn_simconf = sim_conf['NODE_ALLOCATION_MODE']
            if pn_simconf.upper() == 'SHARED':
                sim_data.sim_conf['NODE_ALLOCATION_MODE'] = 'SHARED'
            elif pn_simconf.upper() == 'EXCLUSIVE':
                sim_data.sim_conf['NODE_ALLOCATION_MODE'] = 'EXCLUSIVE'
            else:
                self.fwk.exception("Bad 'NODE_ALLOCATION_MODE' value %s" % pn_simconf)
                raise Exception("Bad 'NODE_ALLOCATION_MODE' value %s" % pn_simconf)
        except Exception:
            sim_data.sim_conf['NODE_ALLOCATION_MODE'] = self.platform_conf['NODE_ALLOCATION_MODE']

        for port in ports_list:
            try:
                comp_ref = ports_config[port]['IMPLEMENTATION']
                if comp_ref.strip() == '':
                    continue
                comp_conf = sim_conf[comp_ref]
            except Exception:
                self.fwk.exception('Error accessing configuration section for ' +
                                   'component %s in simulation %s', comp_ref, sim_name)
                sys.exit(1)
            conf_fields = set(comp_conf.keys())

            # Move the paths to the component levels so that they can use it
            # If they already have it, then they are effectively overriding the global values
            if 'INPUT_DIR' not in comp_conf:
                if 'INPUT_DIR' in sim_conf:
                    comp_conf['INPUT_DIR'] = sim_conf['INPUT_DIR']
                else:
                    comp_conf['INPUT_DIR'] = sim_data.conf_file_dir
            if 'IPS_ROOT' not in comp_conf:
                if 'IPS_ROOT' in sim_conf:
                    comp_conf['IPS_ROOT'] = sim_conf['IPS_ROOT']
            if 'DATA_TREE_ROOT' not in comp_conf:
                if 'DATA_TREE_ROOT' in sim_conf:
                    comp_conf['DATA_TREE_ROOT'] = sim_conf['DATA_TREE_ROOT']
                else:
                    comp_conf['DATA_TREE_ROOT'] = sim_data.conf_file_dir
            if 'BIN_DIR' not in comp_conf:
                if 'BIN_DIR' in sim_conf:
                    comp_conf['BIN_DIR'] = sim_conf['BIN_DIR']
            if 'BIN_PATH' not in comp_conf:
                if 'BIN_PATH' in sim_conf:
                    comp_conf['BIN_PATH'] = sim_conf['BIN_PATH']
                else:
                    comp_conf['BIN_PATH'] = comp_conf['BIN_DIR']
            if not self.required_fields.issubset(conf_fields):
                msg = 'Error: missing required entries {} in simulation {} component {} configuration section'.format(
                    list(self.required_fields - conf_fields), sim_name, comp_ref)
                self.fwk.critical(msg)
                raise RuntimeError(msg)
            component_id = self._create_component(comp_conf, sim_data)
            sim_data.port_map[port] = component_id
            if port == 'DRIVER':
                sim_data.driver_comp = component_id
            elif port == 'INIT':
                sim_data.init_comp = component_id

        if sim_data.driver_comp is None:
            msg = 'Missing DRIVER specification in config file for simulation {}'.format(sim_data.sim_name)
            self.fwk.critical(msg)
            raise RuntimeError(msg)
        if sim_data.init_comp is None:
            self.fwk.warning('Missing INIT specification in ' +
                             'config file for simulation %s', sim_data.sim_name)

    def _create_component(self, comp_conf, sim_data):
        """
        Create component and populate it with the information from the
        component's configuration section.
        """
        sim_name = sim_data.sim_name
        class_name = comp_conf['NAME']

        if comp_conf['SCRIPT']:
            try:
                fullpath = os.path.abspath(comp_conf['SCRIPT'])
                script = comp_conf['SCRIPT'].rsplit('.', 1)[0].split('/')[-1]
                spec = importlib.util.spec_from_file_location(script, fullpath)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                component_class = getattr(module, class_name)
            except (FileNotFoundError, AttributeError):
                self.fwk.error('Error in configuration file : NAME = %s   SCRIPT = %s',
                               comp_conf['NAME'], comp_conf['SCRIPT'])
                self.fwk.exception('Error instantiating IPS component %s From %s', class_name, script)
                raise
        else:
            try:
                module = importlib.import_module(comp_conf['MODULE'])
                component_class = getattr(module, class_name)
            except (ModuleNotFoundError, AttributeError):
                raise

        # SIMYAN: removed else conditional, copying files in runspaceInit
        # component now

        svc_response_q = Queue(0)
        invocation_q = Queue(0)
        component_id = ComponentID(class_name, sim_name)
        fwk_inq = self.fwk.get_inq()

        log_pipe_name = sim_data.log_pipe_name
        services_proxy = ServicesProxy(self.fwk, fwk_inq, svc_response_q,
                                       sim_data.sim_conf, log_pipe_name)
        new_component = component_class(services_proxy, comp_conf)
        new_component.__initialize__(component_id, invocation_q, self.fwk.start_time)
        services_proxy.__initialize__(new_component)
        self.comp_registry.addEntry(component_id, svc_response_q,
                                    invocation_q, new_component,
                                    services_proxy,
                                    comp_conf)
        p = Process(target=new_component.__run__)
        p.start()
        sim_data.process_list.append(p)
        sim_data.all_comps.append(component_id)
        return component_id

    def get_component_map(self):
        """
        Return a dictionary of simulation names and lists of component
        references.  (May only be the driver, and init (if present)???)
        """
        sim_comps = {}
        for sim_name in self.sim_map:
            if sim_name == self.fwk_sim_name:
                continue
            sim_comps[sim_name] = self.get_simulation_components(sim_name)
        return sim_comps

    def get_simulation_components(self, sim_name):
        comp_list = []
        sim_data = self.sim_map[sim_name]
        if sim_data.init_comp:
            comp_list.append(sim_data.init_comp)
        comp_list.append(sim_data.driver_comp)
        return comp_list

    def get_all_simulation_components_map(self):
        sim_comps = {name: self.sim_map[name].all_comps[:]
                     for name in self.sim_map}
        del sim_comps[self.fwk_sim_name]
        return sim_comps

    def get_framework_components(self):
        """
        Return list of framework components.
        """
        fwk_components = self.fwk_components[:]
        return fwk_components

    def get_sim_parameter(self, sim_name, param):
        """
        Return value of *param* from simulation configuration file for
        *sim_name*.
        """
        try:
            sim_data = self.sim_map[sim_name]
        except KeyError:
            sim_data = self.finished_sim_map[sim_name]
        try:
            val = sim_data.sim_conf[param]
        except KeyError:
            val = self.platform_conf[param]
        self.fwk.debug('Returning value = %s for config parameter %s in simulation %s', val, param, sim_name)
        return val

    def get_sim_names(self):
        """
        Return list of names of simulations.
        """
        return list(self.sim_map.keys())

    def process_service_request(self, msg):
        """
        Invokes public configuration manager method for a component.  Return
        method's return value.
        """
        self.fwk.debug('Configuration Manager received message: %s', str(msg.__dict__))
        sim_name = msg.sender_id.get_sim_name()
        method = getattr(self, msg.target_method)
        self.fwk.debug('Configuration manager dispatching method %s on simulation %s',
                       method, sim_name)
        retval = method(sim_name, *msg.args)
        return retval

    def create_simulation(self, sim_name, config_file, override, sub_workflow=False):
        try:
            conf = ConfigObj(config_file, interpolation='template',
                             file_error=True)
        except IOError:
            self.fwk.exception('Error opening config file %s: ', config_file)
            raise
        except SyntaxError:
            self.fwk.exception(' Error parsing config file %s: ', config_file)
            raise
        parent_sim_name = sim_name
        parent_sim = self.sim_map[parent_sim_name]
        # Incorporate environment variables into config file
        # Use config file entries when duplicates are detected
        conf_keys = list(conf.keys())
        for (k, v) in os.environ.items():
            # Do not include functions from environment
            if k not in conf_keys and \
                    not any(x in v for x in '{}()$'):
                conf[k] = v

        # Allow propagation of entries from platform config file to simulation
        # config file
        for keyword in list(self.platform_conf.keys()):
            if keyword not in list(conf.keys()):
                conf[keyword] = self.platform_conf[keyword]
        if override:
            for kw in list(override.keys()):
                conf[kw] = override[kw]
        try:
            sim_name = conf['SIM_NAME']
            sim_root = conf['SIM_ROOT']
            log_file = os.path.abspath(conf['LOG_FILE'])
        except KeyError:
            self.fwk.exception('Missing required parameters SIM_NAME, SIM_ROOT or LOG_FILE\
in configuration file %s', config_file)
            raise
        if sim_name in self.sim_name_list:
            self.fwk.error('Error: Duplicate SIM_NAME %s in configuration files' % (sim_name))
            raise Exception('Duplicate SIM_NAME %s in configuration files' % (sim_name))
        if sim_root in self.sim_root_list:
            self.fwk.exception('Error: Duplicate SIM_ROOT in configuration files')
            raise Exception('Duplicate SIM_ROOT in configuration files')
        if log_file in self.log_file_list:
            self.fwk.exception('Error: Duplicate LOG_FILE in configuration files')
            raise Exception('Duplicate LOG_FILE in configuration files')

        # Add path to configuration file to simulation configuration in memory
        if 'SIMULATION_CONFIG_FILE' not in conf:
            conf['SIMULATION_CONFIG_FILE'] = config_file

        self.sim_name_list.append(sim_name)
        self.sim_root_list.append(sim_root)
        self.log_file_list.append(log_file)
        new_sim = self.SimulationData(sim_name)
        new_sim.sim_conf = conf
        new_sim.config_file = config_file
        new_sim.sim_root = sim_root
        new_sim.log_file = log_file
        if not sub_workflow:
            new_sim.portal_sim_name = sim_name
            new_sim.log_pipe_name = tempfile.mktemp('.logpipe', 'ips_')
            self.log_dynamic_sim_queue.put('CREATE_SIM  %s  %s' % (new_sim.log_pipe_name, new_sim.log_file))
        else:
            new_sim.portal_sim_name = parent_sim.portal_sim_name
            new_sim.log_pipe_name = parent_sim.log_pipe_name

        conf['__PORTAL_SIM_NAME'] = new_sim.portal_sim_name
        self.sim_map[sim_name] = new_sim
        self._initialize_sim(new_sim)

        if not sub_workflow:
            self.fwk.initiate_new_simulation(sim_name)

        return (sim_name, new_sim.init_comp, new_sim.driver_comp)

    def get_port(self, sim_name, port_name):
        """
        Return a reference to the component from simulation *sim_name*
        implementing port *port_name*.
        """
        sim_data = self.sim_map[sim_name]
        comp_id = sim_data.port_map[port_name]
        return comp_id

    def get_config_parameter(self, sim_name, param):
        """
        Return value of *param* from simulation configuration file for
        *sim_name*.
        """
        return self.get_sim_parameter(sim_name, param)

    def set_config_parameter(self, sim_name, param, value, target_sim_name):
        """
        Set the configuration parameter *param* to value *value* in
        *target_sim_name*.  If *target_sim_name* is the framework, all
        simulations will get the change.  Return *value*.
        """
        if target_sim_name == self.fwk_sim_name:  # apply to all simulations
            target_sims = list(self.sim_map.keys())
        else:
            target_sims = [target_sim_name]
        for other_sim_name in target_sims:
            self.fwk.debug('Setting %s to %s in simulation %s', param, value, other_sim_name)
            try:
                sim_data = self.sim_map[other_sim_name]
            except KeyError:
                sim_data = self.finished_sim_map[other_sim_name]
            self.fwk.debug('Setting %s to %s in simulation %s', param, value, other_sim_name)
            sim_conf = sim_data.sim_conf
            sim_conf[param] = value

        return value

    def get_platform_parameter(self, param, silent=False):
        """
        Return value of platform parameter *param*.  If *silent* is ``False``
        (default) ``None`` is returned when *param* not found, otherwise an
        exception is raised.
        """
        val = None
        try:
            val = self.platform_conf[param]
        except KeyError:
            if not silent:
                self.fwk.warning('CM: No platform data for %s ', param)
                raise
        return val

    def terminate_sim(self, sim_name):
        sim_data = self.sim_map[sim_name]
        all_sim_components = sim_data.all_comps
        msg = 'END_SIM %s' % (sim_data.log_pipe_name)
        self.log_dynamic_sim_queue.put(msg)
        proc_list = sim_data.process_list
        for p in proc_list:
            p.terminate()
            p.join()
        try:
            os.remove(sim_data.log_pipe_name)
        except Exception:
            pass
        for comp_id in all_sim_components:
            self.comp_registry.removeEntry(comp_id)
        sim_data.logger = None
        sim_data.process_list = []
        self.finished_sim_map[sim_name] = sim_data
        del self.sim_map[sim_name]

    def terminate(self, status):
        """
        Terminates all processes attached to the framework.  *status* not used.
        """
        try:
            for sim_name in list(self.sim_map.keys()):
                self.terminate_sim(sim_name)
        except Exception:
            print('Encountered exception when terminating simulation')
            raise
        for k in list(self.sim_map.keys()):
            del self.sim_map[k]
        self.log_process.terminate()



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
"""
This file writes debug messages to 'debug.out' file if the environment
variable 'IPSES_DEBUG' is defined.
"""

import os


class Debug:
    def __init__(self):
        self.file = None
        if 'IPSES_DEBUG' in os.environ:
            self.file = open('debug.out', 'w')

    def output(self, s, id1=0, id2=0):
        if self.file:
            tmp = ''
            if id1 != 0:
                """ one subscriber/listener """
                if id2 == 0:
                    tmp += ', id = ' + str(id1)
                else:
                    tmp += ', listenerid = ' + str(id1) + ', subscriberid = ' + str(id2)

            self.file.write(s + tmp + '\n')

    def msg(self, s1, ret=99, s2=''):
        if self.file:
            if s2 == '':
                if ret != 99:
                    self.file.write(s1 + ' ' + str(ret) + '\n')
                else:
                    self.file.write(s1 + '\n')
            else:
                if ret != 99:
                    self.file.write(s1 + ' ' + str(ret) + ' ' + s2 + '\n')
                else:
                    self.file.write(s1 + ' ' + s2 + '\n')

    def __del__(self):
        if self.file:
            self.file.close()


debug = Debug()



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
# local version
import os
import time
from math import ceil
from .ipsExceptions import InsufficientResourcesException, \
    BadResourceRequestException, \
    ResourceRequestMismatchException
from .ips_es_spec import eventManager
from .resourceHelper import getResourceList
from .node_structure import Node


class ResourceManager:
    """
    The resource manager is responsible for detecting the resources allocated
    to the framework, allocating resources to task requests, and maintaining
    the associated bookkeeping.
    """
    # RM init

    def __init__(self, fwk):
        """
        declaration of bookkeeping values
        """
        self.rm_start_of_time = time.time()
        # ref to framework
        self.fwk = fwk
        self.EM = None
        self.DM = None
        self.TM = None
        self.CM = None

        self.accurateNodes = False

        # bookkeeping for allocationa and accounting
        self.nodes = {}
        self.num_nodes = 0
        self.avail_nodes = list()
        self.alloc_nodes = list()
        self.total_cores = 0
        self.alloc_cores = 0
        self.avail_cores = 0
        self.processes = 0
        self.active_tasks = {}  # tid:(owner, cores, procs)

        # hardware node topology
        self.cores_per_node = 1
        self.sockets_per_node = 1
        self.cores_per_socket = 1

        # other stuff
        self.max_ppn = 1   # the ppn for the whole submission (max ppn allowed by *software*)
        self.ppn = 1  # platform config ppn for the whole IPS
        self.myTopic = None
        self.service_methods = ['get_allocation', 'release_allocation']

        self.fwk.register_service_handler(self.service_methods,
                                          getattr(self, 'process_service_request'))
    # RM initialize

    def initialize(self, dataMngr, taskMngr, configMngr,
                   cmd_nodes=0, cmd_ppn=0):
        """
        Initialize resource management structures, references to other
        managers (*dataMngr*, *taskMngr*, *configMngr*).

        Resource information comes from the following in order of priority:

          * command line specification (*cmd_nodes*, *cmd_ppn*)
          * detection using parameters from platform config file
          * manual settings from platform config file

        The second two sources are obtained through
        :py:meth:`resourceHelper.getResourceList`.
        """
        self.EM = eventManager(self)
        self.DM = dataMngr
        self.TM = taskMngr
        self.CM = configMngr
        self.node_alloc_mode = self.CM.get_platform_parameter('NODE_ALLOCATION_MODE')

        rfile_name = os.path.join(self.CM.sim_map[self.CM.fwk_sim_name].sim_root, "resource_usage")
        # SIMYAN: try to safely make the directory...
        try:
            os.makedirs(self.CM.sim_map[self.CM.fwk_sim_name].sim_root)
        except OSError as oserr:
            if oserr.errno != 17:
                raise

        self.reporting_file = open(rfile_name, "w")

        # -------------------------------
        # check cmd line resource spec
        # -------------------------------
        if cmd_nodes != 0 and cmd_ppn != 0:
            # use cmd resource specification
            self.host = "override_%s" % os.environ['HOST']
            self.max_ppn = cmd_ppn
            self.cores_per_node = int(cmd_ppn)
            self.ppn = int(cmd_ppn)
            self.sockets_per_node = 1
            self.accurateNodes = False
            listOfNodes = []
            for i in range(cmd_nodes):
                listOfNodes.append(("dummy_node%d" % i, cmd_ppn))
        else:
            self.ppn = 0
            # -------------------------------
            # get resource information
            # -------------------------------
            self.host = self.CM.get_platform_parameter('HOST')
            self.fwk.debug('RM: Host = %s', self.host)
            try:
                listOfNodes, self.cores_per_node, self.sockets_per_node,  \
                    self.max_ppn, self.accurateNodes = \
                    getResourceList(self.CM, self.host)
                self.fwk.warning('RM: listOfNodes = %s', str(listOfNodes))
                self.fwk.warning('RM: max_ppn = %d ', int(self.max_ppn))
                if self.accurateNodes is True and not self.CM.get_platform_parameter('USE_ACCURATE_NODES'):
                    self.accurateNodes = False
                    self.fwk.warning('RM: User set accurateNodes to False')
            except Exception:
                print("can't get resource info")
                raise

            # -------------------------------
            # set ppn
            # -------------------------------
            uppn_config = int(self.CM.get_platform_parameter('PROCS_PER_NODE'))
            if uppn_config == 0:
                user_ppn = self.max_ppn
            else:
                user_ppn = uppn_config

            if user_ppn <= self.max_ppn:
                self.ppn = user_ppn
                for i, (node, count) in enumerate(listOfNodes):
                    if count > self.ppn:
                        listOfNodes[i] = (node, self.ppn)
                self.fwk.warning("Using user set procs per node: %d", user_ppn)
            else:
                self.fwk.warning("Platform specified  PROCS_PER_NODE = %d is greater than batch job specification = %d." % (user_ppn, self.max_ppn))
                self.fwk.warning("Will use batch job specification to launch tasks")
                self.ppn = self.max_ppn

            try:
                if user_ppn <= self.max_ppn:
                    self.ppn = user_ppn
                    for i, (node, count) in enumerate(listOfNodes):
                        if count > self.ppn:
                            listOfNodes[i] = (node, self.ppn)
                else:
                    self.fwk.warning("Platform specified  PROCS_PER_NODE is greater than batch job specification.")
                    self.fwk.warning("Will use batch job specification to launch tasks")
                    self.ppn = self.max_ppn
            except Exception:
                # this stipulation doesn't make sense for explicitly named cores
                pass

            # -----------------------------------
            # set cores/sockets per socket/node
            # -----------------------------------
            if (self.cores_per_node % self.sockets_per_node) == 0:
                self.cores_per_socket = self.cores_per_node // self.sockets_per_node
            else:
                self.fwk.warning("cpn (%d) not divisible by spn(%d) - setting spn to 1" % (self.cores_per_node, self.sockets_per_node))
                self.sockets_per_node = 1
                self.cores_per_socket = self.cores_per_node

        # -------------------------------
        # populate nodes
        # -------------------------------
        self.fwk.warning('RM: %d nodes and %d processors per node' % (len(listOfNodes), self.ppn))
        self.total_cores = self.add_nodes(listOfNodes)
        self.avail_cores = self.total_cores
        self.begin_RM_report()

    def process_service_request(self, msg):
        pass

    def begin_RM_report(self):
        """
        Print header information for resource usage reporting file.
        """
        print("# host:", self.host, file=self.reporting_file)
        print("# total nodes:", self.num_nodes, file=self.reporting_file)
        print("# processors per node:", self.ppn, file=self.reporting_file)
        print("using accurate nodes:", self.accurateNodes, file=self.reporting_file)
        print("# time (in seconds since the | available | allocated | percent allocated | processes | percent used | notes ", file=self.reporting_file)
        print("#   resource manager started |           |           |                   |           |              |", file=self.reporting_file)
        print("#-----------------------------------------------------------------------------------------------------------", file=self.reporting_file)
        self.report_RM_status('initial state of resources')

    def report_RM_status(self, notes=""):
        """
        Print current RM status to the reporting_file ("resource_usage")
        Entries consist of:

         - time in seconds since beginning of time (__init__ of RM)
         - # cores that are available
         - # cores that are allocated
         - % allocated cores
         - # processes launched by task
         - % cores used by processes
         - notes (a description of the event that changed the resource usage)
        """
        print(" %27.5f |" % (time.time() - self.rm_start_of_time), end=' ', file=self.reporting_file)
        print(" %8d |" % self.avail_cores, end=' ', file=self.reporting_file)
        print(" %8d |" % self.alloc_cores, end=' ', file=self.reporting_file)
        print(" %16.2f |" % (100 * (float(self.alloc_cores) / self.total_cores)), end=' ', file=self.reporting_file)
        print(" %8d |" % self.processes, end=' ', file=self.reporting_file)
        print(" %10.2f  # " % (100 * (float(self.processes) / self.total_cores)), end=' ', file=self.reporting_file)
        print(notes, file=self.reporting_file)
        self.reporting_file.flush()

    def printRMState(self):
        """
        Print the node tree to ``stdout``.
        """
        print("*** RM.nodeTable ***")
        for n, i in list(self.nodes.items()):
            print(n)
            i.print_sockets()
        print("=====================")

    def add_nodes(self, listOfNodes):
        """
        Add node entries to ``self.nodes``.  Typically used by
        :py:meth:`.initialize` to initialize ``self.nodes``.
        May be used to add nodes to a dynamic allocation in the future.

        *listOfNodes* is a list of tuples (*node name*, *cores*).
        ``self.nodes`` is a dictionary where the keys are the *node names* and
        the values are :py:class:`node_structure.Node` structures.

        Return total number of cores.
        """
        tot_cores = 0
        for n, p in listOfNodes:
            if n not in self.nodes:
                self.nodes.update({n: Node(n, self.sockets_per_node,
                                           self.cores_per_node, p)})
                self.num_nodes += 1
                self.avail_nodes.append(n)
                if isinstance(p, int):
                    tot_cores += p
                else:  # p is a list of core names
                    tot_cores += len(p)
        return tot_cores

    # RM getAllocation
    # pylint: disable=inconsistent-return-statements
    def get_allocation(self, comp_id, nproc, task_id,
                       whole_nodes, whole_socks, task_ppn=0):
        """
        Traverse available nodes to return:

        If *whole_nodes* is ``True``:

          * *shared_nodes*: ``False``
          * *nodes*: list of node names
          * *ppn*: processes per node for launching the task
          * *max_ppn*: processes that can be launched
          * *accurateNodes*: ``True`` if *nodes* uses the actual names of the nodes, ``False`` otherwise.

        If *whole_nodes* is ``False``:

          * *shared_nodes*: ``True``
          * *nodes*: list of node names
          * *node_file_entries*: list of (node, corelist) tuples, where *corelist* is a list of core names.
             Core names are integers from 0 to n-1 where n is the number of cores on a node.
          * *ppn*: processes per node for launching the task
          * *max_ppn*: processes that can be launched
          * *accurateNodes*: ``True`` if *nodes* uses the actual names of the nodes, ``False`` otherwise.

        Aguments:

          * *nproc*: the number of requested processes (int)
          * *comp_id*: component identifier, must be unique with respect to the framework (string)
          * *task_id*: task identifier from TM (int)
          * *method*: name of method (string)
          * *task_ppn*: ppn for this task (optional) (int)
        """
        # get the component requirements for all of the components

        # set ppn for this task
        if task_ppn > 0:
            if task_ppn > self.ppn:
                if task_ppn > self.max_ppn:
                    self.fwk.warning("task ppn exceeds machine ppn, using machine ppn instead")
                    ppn = self.max_ppn
                else:
                    ppn = task_ppn
            else:
                ppn = task_ppn
        else:
            ppn = self.ppn

        if nproc < ppn:
            ppn = nproc

        # check if partial node allocation is possible
        if self.node_alloc_mode == "EXCLUSIVE":
            if not (whole_nodes and whole_socks):
                self.fwk.warning("No partial node allocation available on this platform, using whole nodes instead.")
            whole_nodes = True
            whole_socks = True

        # Are there enough cores to satisfy the request?
        # Returns the list of nodes that fit the bill
        allocation_possible = False
        if whole_nodes:
            allocation_possible, nodes = self.check_whole_node_cap(nproc, ppn)
        elif whole_socks:
            allocation_possible, nodes = self.check_whole_sock_cap(nproc, ppn)
        else:
            allocation_possible, nodes = self.check_core_cap(nproc, ppn)

        if not allocation_possible:
            if nodes == "bad":
                c = ceil(float(nproc) / ppn)
                raise BadResourceRequestException(comp_id, task_id, c, c - len(self.avail_nodes))
            if nodes == "mismatch":
                raise ResourceRequestMismatchException(comp_id, task_id,
                                                       nproc, ppn,
                                                       self.total_cores,
                                                       self.max_ppn)
            if nodes == "insufficient":
                c = ceil(float(nproc) / ppn)
                raise InsufficientResourcesException(comp_id, task_id,
                                                     c, c - len(self.avail_nodes))
        else:
            try:
                self.processes += nproc
                k = 0
                alloc_procs = 0
                node_file_entries = []
                if whole_nodes:
                    # -------------------------------
                    # whole node allocation
                    # -------------------------------
                    for n in nodes:
                        procs, cores = self.nodes[n].allocate(whole_nodes,
                                                              whole_socks,
                                                              task_id, comp_id,
                                                              ppn)
                        self.avail_nodes.remove(n)
                        self.alloc_nodes.append(n)
                        node_file_entries.append((n, cores))
                        k += procs
                    self.alloc_cores += k
                    self.avail_cores -= k
                    self.active_tasks.update({task_id: (comp_id, nproc, k)})
                elif whole_socks:
                    # -------------------------------
                    # whole sock allocation
                    # -------------------------------
                    for n in nodes:
                        node = self.nodes[n]
                        if node.avail_cores > 0:
                            to_alloc = min([ppn, node.avail_cores,
                                            nproc - alloc_procs])
                            procs, cores = node.allocate(whole_nodes,
                                                         whole_socks,
                                                         task_id, comp_id,
                                                         to_alloc)
                            k += len(cores)
                            alloc_procs = min([ppn, len(cores)])
                            node_file_entries.append((n, cores))
                            if n not in self.alloc_nodes:
                                self.alloc_nodes.append(n)
                                if node.avail_cores - node.total_cores == 0:
                                    self.avail_nodes.remove(n)

                    self.alloc_cores += k
                    self.avail_cores -= k
                    self.active_tasks.update({task_id: (comp_id, nproc, k)})
                else:
                    # -------------------------------
                    # single core allocation
                    # -------------------------------
                    for n in nodes:
                        node = self.nodes[n]
                        if node.avail_cores > 0:
                            to_alloc = min([ppn, node.avail_cores,
                                            nproc - k])
                            self.fwk.debug("allocate task_id %d node %s %d cores" % (task_id, n, to_alloc))
                            procs, cores = node.allocate(whole_nodes,
                                                         whole_socks,
                                                         task_id, comp_id,
                                                         to_alloc)
                            k += procs
                            node_file_entries.append((n, cores))
                            if n not in self.alloc_nodes:
                                self.alloc_nodes.append(n)
                                if node.avail_cores - node.total_cores == 0:
                                    self.avail_nodes.remove(n)

                    self.alloc_cores += k
                    self.avail_cores -= k
                    self.active_tasks.update({task_id: (comp_id, nproc, k)})
            except Exception:
                print("Available Nodes:")
                for nm in self.avail_nodes:
                    n = self.nodes[nm]
                    print(n.name, n.avail_cores)
                    n.print_sockets()
                print("\nAllocated Nodes:")
                for nm in self.alloc_nodes:
                    n = self.nodes[nm]
                    print(n.name, n.avail_cores)
                    n.print_sockets()
                print("\n ***** Neither List!")
                for nm, n in list(self.nodes.items()):
                    if nm not in self.avail_nodes and nm not in self.alloc_nodes:
                        print(nm, n.avail_cores)
                        n.print_sockets()
                raise

            if whole_nodes:
                self.report_RM_status("allocation for task %d using whole nodes" % task_id)
                return not whole_nodes, nodes, ppn, self.max_ppn, self.accurateNodes
            else:
                self.report_RM_status("allocation for task %d using partial nodes" % task_id)
                return not whole_nodes, nodes, node_file_entries, ppn, self.max_ppn, self.accurateNodes

    def check_whole_node_cap(self, nproc, ppn):
        """
        Determine if it is currently possible to allocate *nproc* processes
        with a ppn of *ppn* and whole nodes.  Return ``True`` and list of
        nodes to use if successful.  Return ``False`` and empty list if there
        are not enough available resources at this time, but it is possible to
        eventually satisfy the request.  Exception raised if the request can
        never be fulfilled.
        """
        whole_cap = 0
        nodes = []
        try:
            for n in self.avail_nodes:
                node = self.nodes[n]
                if node.avail_cores == node.total_cores and node.avail_cores >= ppn:
                    whole_cap += ppn
                    nodes.append(n)
                    if whole_cap >= nproc:
                        return True, nodes
        except Exception:
            self.fwk.exception("problem in RM.check_whole_node_cap")
            raise
        # check to see if it is possible to satify the request
        tot_cap = 0
        for n in list(self.nodes.values()):
            tot_cap += min([ppn, n.total_cores])
            if tot_cap >= nproc:
                return False, "insufficient"

        if self.total_cores < nproc:
            return False, "bad"
        else:
            return False, "mismatch"

    def check_whole_sock_cap(self, nproc, ppn):
        """
        Determine if it is currently possible to allocate *nproc* processes
        with a ppn of *ppn* and whole sockets.  Return ``True`` and list of
        nodes to use if successful.  Return ``False`` and empty list if there
        are not enough available resources at this time, but it is possible to
        eventually satisfy the request.  Exception raised if the request can
        never be fulfilled.
        """
        nodes = []
        k = 0
        try:
            for n in self.avail_nodes:
                node = self.nodes[n]
                sk = 0
                for sock in node.sockets:
                    if sock.total_cores == sock.avail_cores:
                        # whole socket
                        if n not in nodes:
                            nodes.append(n)
                        if sock.avail_cores > ppn:
                            sk += ppn
                        else:
                            sk += sock.avail_cores
                        if sk >= ppn:
                            break
                k += sk
                if k >= nproc:
                    return True, nodes
        except Exception:
            self.fwk.exception("problem in RM.check_whole_sock_cap")
            raise
        # check to see if it is possible to satify the request
        tot_cap = 0
        for n in list(self.nodes.values()):
            tot_cap += min([ppn, n.total_cores])
            if tot_cap >= nproc:
                return False, "insufficient"

        if self.total_cores < nproc:
            return False, "bad"
        else:
            return False, "mismatch"

    def check_core_cap(self, nproc, ppn):
        """
        Determine if it is currently possible to allocate *nproc* processes
        with a ppn of *ppn* without further restrictions..  Return ``True``
        and list of nodes to use if successful.  Return ``False`` and empty
        list if there are not enough available resources at this time, but it
        is possible to eventually satisfy the request.  Exception raised if
        the request can never be fulfilled.
        """
        nodes = []
        k = 0
        try:
            for n in self.avail_nodes:
                node = self.nodes[n]
                if nproc - k < ppn:
                    if node.avail_cores >= nproc - k:
                        k = nproc
                        nodes.append(n)
                        self.fwk.debug("found nodes (%s) and returning" % nodes)
                        return True, nodes
                    elif node.avail_cores > 0:
                        k += node.avail_cores
                        nodes.append(n)
                else:
                    if node.avail_cores >= ppn:
                        k += ppn
                        nodes.append(n)
                    elif node.avail_cores > 0:
                        k += node.avail_cores
                        nodes.append(n)
                if k >= nproc:
                    return True, nodes
        except Exception:
            self.fwk.exception("problem in RM.check_core_cap")
            raise
        # check to see if it is possible to satify the request
        tot_cap = 0
        for n in list(self.nodes.values()):
            tot_cap += min([ppn, n.total_cores])
            if tot_cap >= nproc:
                return False, "insufficient"

        if self.total_cores < nproc:
            return False, "bad"
        else:
            return False, "mismatch"

    # RM releaseAllocation
    def release_allocation(self, task_id, status):
        """
        Set resources allocated to task *task_id* to available.  *status* is
        not used, but may be used to correlate resource failures to task
        failures and implement task relaunch strategies.
        """

        o, nproc, num_cores = self.active_tasks[task_id]
        tot_avc = 0
        for n, node in list(self.nodes.items()):
            if task_id in node.task_ids:
                node.release(task_id, o)
                if node.avail_cores > 0 and n not in self.avail_nodes:
                    self.avail_nodes.append(n)
                if node.avail_cores == node.total_cores:
                    self.alloc_nodes.remove(n)
            tot_avc += node.avail_cores

        self.avail_cores += num_cores
        self.alloc_cores -= num_cores
        self.processes -= nproc

        self.report_RM_status('released nodes for task %d' % task_id)

        return True

    # RM SendEvent
    def sendEvent(self, eventName, info):
        """
        wrapper for constructing and publishing EM events
        """
        # -------------------------------
        #     send an event
        # -------------------------------
        # populate event body
        eventBody = {}
        eventBody.update({'event name': eventName, 'topic': 'test', 'sender': 'RM',
                          'data': 'A resource event has occured'})
        eventBody.update(info)
        # send event on topic



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
import re

import datetime
import sys
import os
from subprocess import Popen, PIPE
import time
import inspect
from collections import defaultdict
import hashlib
import glob
import itertools
import json
import shutil
from ipsframework import ipsutil, Component
from ipsframework.convert_log_function import convert_logdata_to_html

try:
    from mpo_arg import mpo_methods as mpo
except ImportError:
    pass
mpo_cert = '/home/elwasif/Projects/atom/MPO/MPO Demo User.pem'
mpo_api = 'https://mpo.psfc.mit.edu/test-api'


def configure_mpo():
    # Use this if you want to include modules from a subfolder or relative path.
    cmd_subfolder = os.path.realpath(os.path.abspath(os.path.join(os.path.split(
        inspect.getfile(inspect.currentframe()))[0], "/home/elwasif/Projects/atom/MPO/client/python")))
    if cmd_subfolder not in sys.path:
        sys.path.insert(0, cmd_subfolder)


def hash_file(file_name):
    '''
    Return the MD5 hash of a file
    :rtype: str
    :param file_name: Full path to file
    :return: MD5 of file_name
    '''
    BLOCKSIZE = 65536
    hasher = hashlib.md5()
    with open(file_name, 'rb') as afile:
        buf = afile.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(BLOCKSIZE)
    return hasher.hexdigest()


class PortalBridge(Component):
    """
    Framework component to communicate with the SWIM web portal.
    """

    class SimulationData:
        """
        Container for simulation data.
        """

        def __init__(self):
            self.counter = 0
            self.monitor_file_name = ""
            self.portal_runid = None
            self.sim_name = ''
            self.sim_root = ''
            self.monitor_file = None
            self.json_monitor_file = None
            self.phys_time_stamp = -1
            self.monitor_url = None
            self.mpo_steps = [None]
            self.mpo_wid = None
            self.bigbuf = ""

    def __init__(self, services, config):
        """
        Declaration of private variables and initialization of
        :py:class:`component.Component` object.
        """
        Component.__init__(self, services, config)
        self.host = ''
        self.curTime = time.localtime()
        self.startTime = self.curTime
        self.services = services
        self.sim_map = {}
        self.portal_url = None
        self.done = False
        self.first_event = True
        self.childProcess = None
        self.mpo = None
        self.mpo_name_counter = defaultdict(lambda: 0)
        self.file_hash_cache = defaultdict(dict)
        self.file_uid_cache = defaultdict(dict)
        self.counter = 0
        self.dump_freq = 10
        self.min_dump_interval = 300  # Minimum time interval in Sec for HTML dump operation
        self.last_dump_time = time.time()
        self.write_to_htmldir = True
        self.html_dir = ""

    def init(self, timestamp=0.0, **keywords):
        """
        Try to connect to the portal, subscribe to *_IPS_MONITOR* events and
        register callback :py:meth:`.process_event`.
        """
        try:
            self.portal_url = self.PORTAL_URL
        except AttributeError:
            pass
        self.host = self.services.get_config_param('HOST')
        self.services.subscribe('_IPS_MONITOR', "process_event")
        try:
            freq = int(self.services.get_config_param("HTML_DUMP_FREQ", silent=True))
        except Exception:
            pass
        else:
            self.dump_freq = freq

        try:
            ENABLE_MPO = os.environ['ENABLE_MPO']
        except KeyError:
            ENABLE_MPO = False
        else:
            ENABLE_MPO = True
        if ENABLE_MPO:
            configure_mpo()
            try:
                self.mpo = mpo(api_url=mpo_api, cert=mpo_cert, debug=True)
                self.mpo.debug = False
                self.mpo.filter = 'json'
            except NameError as e:
                print("#################", e)

        try:
            self.html_dir = self.services.get_config_param("USER_W3_DIR", silent=True)
        except Exception:
            self.services.warning("Missing USER_W3_DIR configuration - disabling web-visible logging")
            self.write_to_htmldir = False
        else:
            if self.html_dir.strip() == '':
                self.services.warning("Empty USER_W3_DIR configuration - disabling web-visible logging")
                self.write_to_htmldir = False
            else:
                try:
                    os.mkdir(self.html_dir)
                except FileExistsError:
                    pass
                except Exception:
                    self.services.warning("Unable to create HTML directory - disabling web-visible logging")
                    self.write_to_htmldir = False

    def step(self, timestamp=0.0, **keywords):
        """
        Poll for events.
        """
        while not self.done:
            self.services.process_events()
            time.sleep(0.5)

    def finalize(self, timestamp=0.0, **keywords):
        for sim_name in list(self.sim_map.keys()):
            sim_data = self.sim_map[sim_name]
            try:
                sim_data.monitor_file.close()
                sim_data.json_monitor_file.close()
            except Exception:
                pass

    def get_elapsed_time(self):
        """
        Return total elapsed time since simulation started in seconds
        (including a possible fraction)
        """
        self.curTime = time.localtime()
        delta_t = time.mktime(self.curTime) - time.mktime(self.startTime)
        return delta_t

    def process_event(self, topicName, theEvent):
        """
        Process a single event *theEvent* on topic *topicName*.
        """
        event_body = theEvent.getBody()
        sim_name = event_body['sim_name']
        portal_data = event_body['portal_data']
        try:
            portal_data["sim_name"] = event_body['real_sim_name']
        except KeyError:
            portal_data["sim_name"] = sim_name

        if portal_data['eventtype'] == 'IPS_START':
            sim_root = event_body['sim_root']
            self.init_simulation(sim_name, sim_root)

        sim_data = self.sim_map[sim_name]
        if portal_data['eventtype'] == 'PORTALBRIDGE_UPDATE_TIMESTAMP':
            sim_data.phys_time_stamp = portal_data['phystimestamp']
            return
        else:
            portal_data['phystimestamp'] = sim_data.phys_time_stamp

        if portal_data['eventtype'] == 'IPS_SET_MONITOR_URL':
            sim_data.monitor_url = portal_data['vizurl']
        elif sim_data.monitor_url:
            portal_data['vizurl'] = sim_data.monitor_url

        #        portal_data['walltime'] = int(self.get_elapsed_time())
        portal_data['portal_runid'] = sim_data.portal_runid
        portal_data['seqnum'] = sim_data.counter

        self.send_event(sim_data, portal_data)
        sim_data.counter += 1
        self.counter += 1

        if portal_data['eventtype'] == 'IPS_END':
            del self.sim_map[sim_name]

        if len(self.sim_map) == 0:
            if self.childProcess:
                self.childProcess.stdin.close()
                self.childProcess.wait()
            self.done = True
            self.services.debug('No more simulation to monitor - exiting')
            time.sleep(1)
        return

    def send_event(self, sim_data, event_data):
        """
        Send contents of *event_data* and *sim_data* to portal.
        """
        timestamp = ipsutil.getTimeString()
        buf = '%8d %s ' % (sim_data.counter, timestamp)
        for (k, v) in event_data.items():
            if len(str(v).strip()) == 0:
                continue
            if ' ' in str(v):
                buf += '%s=\'%s\' ' % (k, str(v))
            else:
                buf += '%s=%s ' % (k, str(v))
        buf += '\n'
        sim_data.monitor_file.write(bytes(buf, encoding='UTF-8'))
        sim_data.bigbuf += buf

        buf = json.dumps(event_data)
        sim_data.json_monitor_file.write("%s\n" % buf)

        freq = self.dump_freq
        if (((self.counter % freq == 0) and (time.time() - self.last_dump_time > self.min_dump_interval)) or
                (event_data['eventtype'] == 'IPS_END')):
            self.last_dump_time = time.time()
            html_filename = sim_data.monitor_file_name.replace('eventlog', 'html')
            html_page = convert_logdata_to_html(sim_data.bigbuf)
            open(html_filename, "w").writelines(html_page)
            if self.write_to_htmldir:
                html_file = os.path.join(self.html_dir, os.path.basename(html_filename))
                try:
                    open(html_file, "w").writelines(html_page)
                except Exception:
                    self.services.exception("Error writing html file into USER_W3_DIR directory")
                    self.write_to_htmldir = False
        if self.portal_url:
            webmsg = json.dumps(event_data)
            try:
                if self.first_event:  # First time, launch sendPost.py daemon
                    cmd = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'sendPost.py')
                    python_exec = shutil.which('python')
                    self.childProcess = Popen([python_exec, cmd], bufsize=128,
                                              stdin=PIPE, stdout=PIPE,
                                              stderr=PIPE, close_fds=True)
                    self.first_event = False
                self.childProcess.stdin.write(('%s %s\n' %
                                               (self.portal_url, webmsg)).encode())
                self.childProcess.stdin.flush()
            except Exception as e:
                self.services.exception('Error transmitting event number %6d to %s : %s',
                                        sim_data.counter, self.portal_url, str(e))
        if sim_data.mpo_wid:
            self.send_mpo_data(event_data, sim_data)

    def send_mpo_data(self, event_data, sim_data):
        def md5(fname):
            "Courtesy of stackoverflow 3431825"
            hash_md5 = hashlib.md5()
            with open(fname, "rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    hash_md5.update(chunk)
            return hash_md5.hexdigest()

        def mpo_add_file(workflow, parent, file, shortname='Need a name', longdesc='did not add a description.'):
            """Add a local file to the workflow attaching to parent. Calculate
            checksum and if the file is already in the mpo database, use the
            already the UID of the already existing file when adding the data
            object - this creates a linkage to the original. The checksum and
            local file path and name are added as metadata.

            This function relies on the user space metadata, ips_checksum
            and ips_filename. The checksum is the md5 sum and the filename
            is expected should have at least a relative qualifying path.

            workflow : workflow_id
            parent : parent_id
            """
            # if file exist, look for its checksum in the database
            try:
                checksum = md5(file)
            except Exception:
                print(("checksum could not find file:", file))
                raise

            is_checksum = self.mpo.search('metadata', params={'key': 'ips_checksum', 'value': checksum})
            # search always returns a list of dictionaries
            # if checksum exists, use first dataobject that has it
            # api search results are sorted by time
            # Note, check this with eqdsk dataobject in test-api
            print('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
            print(len(is_checksum), file, parent)
            print('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')

            if len(is_checksum) > 0:
                # uid is chosen to be first occurrence
                # parent_uid is uid of object metadata is attached to.
                file_uid = is_checksum[0]['parent_uid']

                # Create dataobject reference by uid in the workflow
                dataobject = self.mpo.add(workflow, parent, uid=file_uid, name=shortname, desc=longdesc)
                print('Existing dataobject : id = ', file_uid)
                self.mpo.meta(dataobject['uid'], 'ips_checksum', checksum)
                # add filename metadata the dataobject reference
                self.mpo.meta(dataobject['uid'], 'ips_filename', file)
            else:
                print(('file', file))
                file_uri = file
                # Create new dataobject by uri and insert reference in to workflow
                dataobject = self.mpo.add(workflow, parent, uri=file_uri, name=shortname, desc=longdesc)
                # add checksum metadata to original data object
                # add function currently only returns uri field, so fetch full record
                full_dataobject = self.mpo.search('dataobject/' + dataobject['uid'])[0]
                print('New data object', full_dataobject['do_uid'])
                # add checksum so dataobject and also
                self.mpo.meta(full_dataobject['do_uid'], 'ips_checksum', checksum)
                self.mpo.meta(dataobject['uid'], 'ips_filename', file)
                self.mpo.meta(dataobject['uid'], 'ips_checksum', checksum)
                dataobject = full_dataobject
            return dataobject

        """
        def get_file_uid(path, file_name):
            '''
            Get file_name's cached MPO uid if seen before in simulation
            :param file_name: Full path to file
            :return: tuple (MPO uid if found, None if not found, file_hash)
            '''
            cached_hash_dict = None
            full_path = os.path.join(path, file_name)
            md5 = hash_file(full_path)
            cached_hash_dict = self.file_hash_cache[file_name]
            try:
                cached_hash_dict[md5]
            except KeyError:
                pass
            return (None, md5)  # (f_uid, hash)
        """

        recordable_events = ['IPS_CALL_BEGIN', 'IPS_STAGE_INPUTS', 'IPS_STAGE_OUTPUTS', 'IPS_CALL_END']
        recordable_mpo_activities = ['IPS_CALL_BEGIN']
        comment = event_data['comment']
        event_type = event_data['eventtype']

        if event_type not in recordable_events:
            return
        inp_objs = []
        if event_type == 'IPS_CALL_END':
            del sim_data.mpo_steps[-1]
            return
        try:
            if event_type == 'IPS_STAGE_INPUTS':
                r = re.compile(r"^Elapsed time = ([0-9]*\.[0-9]*) Path = ([^ ]*) Files = (.*)")
                o = r.match(comment)
                (_, path, files) = o.groups()
                glist = [glob.glob(os.path.join(path, f)) for f in files.split()]
                for file_name in [os.path.basename(f) for f in itertools.chain(*glist)]:
                    """
                    (f_uid, f_hash) = get_file_uid(path, file_name)
                    f_uid = False
                    if f_uid:
                        mpo_data_obj = self.mpo.add(workflow_ID=sim_data.mpo_wid,
                                                    parentobj_ID=sim_data.mpo_wid['uid'],
                                                    name=file_name,
                                                    desc="An input file",
                                                    uri='file:' + os.path.join(path, file_name),
                                                    uid=f_uid)
                    else:
                        mpo_data_obj = self.mpo.add(workflow_ID=sim_data.mpo_wid,
                                                    parentobj_ID=sim_data.mpo_wid['uid'],
                                                    name=file_name,
                                                    desc="An input file",
                                                    uri='file:' + os.path.join(path, file_name))
                    """
                    mpo_data_obj = mpo_add_file(sim_data.mpo_wid,
                                                sim_data.mpo_wid['uid'],
                                                os.path.join(path, file_name),
                                                shortname=file_name,
                                                longdesc="An input file")
                    inp_objs.append(mpo_data_obj['uid'])

            if event_type == 'IPS_STAGE_INPUTS' and not inp_objs:
                return

            count = self.mpo_name_counter[sim_data.sim_name + event_data['code']]
            if event_type == 'IPS_CALL_BEGIN':
                target = event_data['comment'].split()[-1]
                step_name = "%s %d" % (target, count)
            else:
                step_name = "{0:s} {1:s} {2:d}" \
                    .format(event_data['code'].split('_')[-1], event_type, count)

            if event_type == 'IPS_STAGE_OUTPUTS':
                r = re.compile(r"^Elapsed time = ([0-9]*\.[0-9]*) Path = ([^ ]*) Files = (.*)")
                o = r.match(comment)
                (_, path, files) = o.groups()
                if not files:
                    return
            activity = self.mpo.step(workflow_ID=sim_data.mpo_wid,
                                     parentobj_ID=sim_data.mpo_steps[-1],
                                     input_objs=inp_objs,
                                     name=step_name,
                                     desc="%s" % event_data['comment'])
            self.mpo_name_counter[sim_data.sim_name + event_data['code']] += 1
            if event_type == 'IPS_STAGE_OUTPUTS':
                glist = [glob.glob(os.path.join(path, f)) for f in files.split()]
                for file_name in [os.path.basename(f) for f in itertools.chain(*glist)]:
                    """
                    (f_uid, f_hash) = get_file_uid(path, file_name)
                    if f_uid:
                        mpo_data_obj = self.mpo.add(workflow_ID=sim_data.mpo_wid,
                                                parentobj_ID=activity['uid'],
                                                name=file_name,
                                                desc="An output file",
                                                uri='file:' + os.path.join(path, file_name),
                                                uid=f_uid,
                                                source = f_uid)
                    else:
                        mpo_data_obj = self.mpo.add(workflow_ID=sim_data.mpo_wid,
                                                parentobj_ID=activity['uid'],
                                                name=file_name,
                                                desc="An output file",
                                                uri='file:' + os.path.join(path, file_name))
                    """
                    mpo_data_obj = mpo_add_file(sim_data.mpo_wid,
                                                activity['uid'],
                                                # sim_data.mpo_wid['uid'],
                                                os.path.join(path, file_name),
                                                shortname=file_name,
                                                longdesc="An output file")

        except Exception as e:
            print("*************", e)
        else:
            if event_type in recordable_mpo_activities:
                sim_data.mpo_steps.append(activity['uid'])

    def init_simulation(self, sim_name, sim_root):
        """
        Create and send information about simulation *sim_name* living in
        *sim_root* so the portal can set up corresponding structures to manage
        data from the sim.
        """
        self.services.debug('Initializing simulation : %s -- %s ', sim_name, sim_root)
        sim_data = self.SimulationData()
        sim_data.sim_name = sim_name
        sim_data.sim_root = sim_root

        d = datetime.datetime.now()
        date_str = "%s.%03d" % (d.strftime("%Y-%m-%dT%H:%M:%S"), int(d.microsecond / 1000))
        sim_data.portal_runid = "_".join([self.host, "USER", date_str])
        try:
            self.services.set_config_param('PORTAL_RUNID', sim_data.portal_runid,
                                           target_sim_name=sim_name)
        except Exception:
            self.services.error('Simulation %s is not accessible', sim_name)
            return

        if sim_data.sim_root.strip() == '.':
            sim_data.sim_root = os.environ['IPS_INITIAL_CWD']
        sim_log_dir = os.path.join(sim_data.sim_root, 'simulation_log')
        try:
            os.makedirs(sim_log_dir)
        except OSError as oserr:
            if oserr.errno != 17:
                self.services.exception('Error creating Simulation Log directory %s : %d %s' %
                                        (sim_log_dir, oserr.errno, oserr.strerror))
                raise

        sim_data.monitor_file_name = os.path.join(sim_log_dir,
                                                  sim_data.sim_name + '_' + sim_data.portal_runid + '.eventlog')
        try:
            sim_data.monitor_file = open(sim_data.monitor_file_name, 'wb', 0)
        except IOError as oserr:
            self.services.error("Error opening file %s: error(%s): %s" %
                                (sim_data.monitor_file_name, oserr.errno, oserr.strerror))
            self.services.error('Using /dev/null instead')
            sim_data.monitor_file = open('/dev/null', 'w')
        json_fname = sim_data.monitor_file_name.replace('eventlog', 'json')
        sim_data.json_monitor_file = open(json_fname, 'w')

        if self.mpo:
            try:
                sim_data.mpo_wid = self.mpo.init(name="SWIM Workflow " + os.environ["USER"],
                                                 desc=sim_data.sim_name,
                                                 wtype="SWIM")
                print("sim_data.mpo_wid = ", sim_data.mpo_wid)
            except Exception as e:
                print(e)
                print("sim_data.mpo_wid = ", sim_data.mpo_wid)
                sim_data.mpo_wid = None
            else:
                sim_data.mpo_steps = [sim_data.mpo_wid['uid']]

        self.sim_map[sim_data.sim_name] = sim_data
        return



#!/usr/bin/env python3
from setuptools import setup, find_packages
import versioneer

long_description = """# Integrated Plasma Simulator (IPS) Framework

The IPS was originally developed for the SWIM project and is designed
for coupling plasma physics codes to simulate the interactions of
various heating methods on plasmas in a tokamak. The physics goal of
the project is to better understand how the heating changes the
properties of the plasma and how these heating methods can be used to
improve the stability of plasmas for fusion energy production.

The IPS framework is thus designed to couple standalone codes flexibly
and easily using python wrappers and file-based data coupling. These
activities are not inherently plasma physics related and the IPS
framework can be considered a general code coupling framework. The
framework provides services to manage:

 - the orchestration of the simulation through component invocation,
    task launch and asynchronous event notification mechanisms
 - configuration of complex simulations using familiar syntax
 - file communication mechanisms for shared and internal (to a
    component) data, as well as checkpoint and restart capabilities

The framework performs the task, configuration, file and resource
management, along with the event service, to provide these features.
"""

setup(
    name="ipsframework",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    url="https://ips-framework.readthedocs.io",
    description="Integrated Plasma Simulator (IPS) Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'ips.py = ipsframework.ips:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)



"""IPS Framework"""
from .component import Component
from .configurationManager import ConfigurationManager
from .dataManager import DataManager
from .ips import Framework
from .resourceManager import ResourceManager
from .services import ServicesProxy, TaskPool, Task
from .taskManager import TaskManager

__all__ = ['Component',
           'ConfigurationManager',
           'DataManager',
           'Framework',
           'ResourceManager',
           'ServicesProxy',
           'TaskPool',
           'Task',
           'TaskManager']

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
from .cca_es_spec import EventListener, PublisherEventService, SubscriberEventService


class myEventListener(EventListener):
    def __init__(self, callback_method):
        super().__init__()
        self.callback_method = callback_method

    def processEvent(self, topicName, theEvent):
        self.callback_method(topicName, theEvent)


class eventManager:
    def __init__(self, obj_ref):
        self.obj_ref = obj_ref
        self.objcache = {}
        self.publisher = "self.publisher"
        self.subscriber = "self.subscriber"

    def publish(self, topicName, eventName, eventBody):
        if self.publisher in self.objcache:
            pub = self.objcache[self.publisher]
        else:
            pub = PublisherEventService()
            self.objcache[self.publisher] = pub

        topic = pub.getTopic(topicName)
        topic.sendEvent(eventName, eventBody)

    def subscribe(self, topicName, callback):
        if not callable(callback):
            callback_method = getattr(self.obj_ref, callback, None)
        else:
            callback_method = callback

        if not callback_method:
            # TODO: do we notify the client that the callback is invalid?
            #      throw an exception?
            return

        if self.subscriber in self.objcache:
            sub = self.objcache[self.subscriber]
        else:
            sub = SubscriberEventService()
            self.objcache[self.subscriber] = sub

        if topicName in self.objcache:
            # TODO: do we notify the client to do an unsubscribe before
            #      re-subscribing to the same topic? the event service
            #      currently throws an exception in this scenario...
            scp = self.objcache[topicName]
        else:
            scp = sub.getSubscription(topicName)
            self.objcache[topicName] = scp

        evl = myEventListener(callback_method)
        scp.registerEventListener(topicName, evl)

    def unsubscribe(self, topicName):
        if topicName in self.objcache:
            self.objcache[topicName].unregisterEventListener(topicName)
            del self.objcache[topicName]
        # else:
            # TODO: do we notify the client to do a subscribe first?
            #      throw an exception?

    def process_events(self):
        if self.subscriber in self.objcache:
            self.objcache[self.subscriber].processEvents()
        # else:
            # TODO: do we notify the client to do a subscribe before processing?
            #      throw an exception?



#!/usr/bin/env python
# -------------------------------------------------------------------------------
#  Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------

import sys
from urllib import request, error
import socket
import time
import traceback

headers = {'Content-Type': 'application/json'}


def sendEncodedMessage(url, msg):
    if not isinstance(msg, bytes):
        msg = msg.encode('utf-8')

    num_trials = 2
    trial = 0
    delay = [0.4, 0.8, 1.2]

    while trial < num_trials:
        try:
            req = request.Request(url, data=msg, headers=headers, method='POST')
            resp = request.urlopen(req)
        except error.URLError:
            trial += 1
            if trial > num_trials:
                open('PORTAL.err', 'a').write('%s\n' % (msg))
            else:
                time.sleep(delay[trial-1])
        else:
            break
    try:
        resp.close()
    except Exception:
        pass


if __name__ == "__main__":
    """ Loop over input from stdin, expecting lines of the format:
        URL ENCODED_WEB_MSG
    """
    timeout = 3
    socket.setdefaulttimeout(timeout)
    error_f = open("sendpost.err", 'w')
    line = '   '
    while True:
        try:
            line = sys.stdin.readline().rstrip('\n')
            if line == '':
                break
            tokens = line.split(' ', 1)
            sendEncodedMessage(tokens[0], tokens[1])
        except Exception:
            traceback.print_exc(file=error_f)
    sys.exit(0)



"""
This file implements several objects that customize logging in the IPS.
"""

import logging
import sys
import pickle
import socketserver
import struct
import functools
import os
import os.path
import queue
import errno
import time
import select


def list_fds():
    """List process currently open FDs and their target """
    ret = {}
    base = '/proc/self/fd'
    for num in os.listdir(base):
        path = None
        try:
            path = os.readlink(os.path.join(base, num))
        except OSError as err:
            # Last FD is always the "listdir" one (which may be closed)
            if err.errno != errno.ENOENT:
                raise
        ret[int(num)] = path

    return ret


class myLogRecordStreamHandler(socketserver.StreamRequestHandler):

    def __init__(self, request, client_address, server, handler):
        self.handler = handler
        socketserver.StreamRequestHandler.__init__(self,
                                                   request,
                                                   client_address,
                                                   server)

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack(">L", chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = self.unPickle(chunk)
            record = logging.makeLogRecord(obj)
            self.handleLogRecord(record)

    def unPickle(self, data):
        return pickle.loads(data)

    def handleLogRecord(self, record):
        name = record.name
        logger = logging.getLogger(name)
        # Need to make sure we only have one handler, since the handler on the
        # sender side is packaged during pickling.
        logger.handlers = []
        logger.addHandler(self.handler)
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        logger.handle(record)
        logger.removeHandler(self.handler)


class LogRecordSocketReceiver(socketserver.ThreadingUnixStreamServer):
    """simple TCP socket-based logging receiver suitable for testing.
    """

    allow_reuse_address = True

    def __init__(self, log_pipe,
                 handler=myLogRecordStreamHandler):
        socketserver.UnixStreamServer.__init__(self, log_pipe, handler)

    def get_file_no(self):
        return self.socket.fileno()


class ipsLogger:
    def __init__(self, dynamic_sim_queue=None):
        self.log_map = {}
        self.server_map = {}
        self.stdout_handler = logging.StreamHandler(sys.stdout)
        self.formatter = logging.Formatter("%(asctime)s %(name)-15s %(levelname)-8s %(message)s")
        self.log_dynamic_sim_queue = dynamic_sim_queue

    def add_sim_log(self, log_pipe_name, log_file=sys.stdout):
        if (log_file == sys.stdout or log_file is None):
            log_handler = self.stdout_handler
        else:
            if log_file.__class__.__name__ == 'TextIOWrapper':
                log_handler = logging.StreamHandler(log_file)
            else:
                directory = os.path.dirname(os.path.abspath(log_file))
                try:
                    os.makedirs(directory)
                except OSError as oserr:
                    if oserr.errno != 17:
                        print('Error creating directory %s : %s-%s' %
                              (directory, oserr.errno, oserr.strerror))
                        sys.exit(1)
                log_handler = logging.FileHandler(log_file, mode='w')

        log_handler.setFormatter(self.formatter)
        partial_handler = functools.partial(myLogRecordStreamHandler,
                                            handler=log_handler)
        recvr = LogRecordSocketReceiver(log_pipe_name, handler=partial_handler)
        fileno = recvr.get_file_no()
        self.log_map[fileno] = (recvr, log_handler, log_pipe_name)

    def __run__(self):
        time_out = 1.0
        while 1:
            read_set = list(self.log_map.keys())
            if read_set:
                rd, _, _ = select.select(read_set,
                                         [], [],
                                         time_out)
            else:
                time.sleep(time_out)
                rd = []

            if len(rd) > 0:
                for fileno in rd:
                    (recvr, _, log_pipe_name) = self.log_map[fileno]
                    recvr.handle_request()
                rd = []
            try:
                msg = self.log_dynamic_sim_queue.get(block=False)
            except queue.Empty:
                pass
            else:
                tokens = msg.split()
                if tokens[0] == 'CREATE_SIM':  # Expecting Message: 'CREATE_SIM  log_pipe_name  log_file
                    self.add_sim_log(tokens[1], tokens[2])
                elif tokens[0] == 'END_SIM':  # Expecting Message 'END_SIM log_pipe_name'
                    log_pipe_name = tokens[1]
                    for fileno, (recvr, _, f_name) in list(self.log_map.items()):
                        if f_name == log_pipe_name:
                            del recvr
                            del self.log_map[fileno]
